BLACK_LIST_DOMAINS = {
    "cdnjs.cloudflare.com",
    "ajax.googleapis.com",
    "maxcdn.bootstrapcdn.com",
    "cdn.statuspage.io",
    "static.getclicky.com"
}

BLACK_LIST_TLD = {
    "cloudfront.net",
    "etclicky.com",
    "eurolandir.com",
    "trafficmanager.net",
    "googleapis.com",
    "gstatic.com",
    "facebook.com",
    "elb.amazonaws.com",
    "amakaiedge.net",
    "googlehosted.com",
    "edgekey.net",
    "s.w.org",
    "slideshare.net",
    "linkedin.com",
    "omniture.com",
    "2O7.net",
    "twitter.com",
    "facebook.com",
    "google.com",
    "youtube.com",
    "googletagmanager.com",
    "wordpress.org",
    "wordpress.com",
    "ghs.googlehosted.com",
    "itunes.apple.com",
    "instagram.com",
    "spotify.com",
    "akadns.net",
    "fontawesome.com",
    "snapchat.com",
    "vimeo.com",
    "gmpg.org",
    "office365.com",
    "office.com",
    "jquery.org",
    "adobe.com",
    "adobelogin.com",
    "stackoverflow.com"
    "jqueryui.com",
    "jquerymobile.com",
    "adform.net",
    "packtpub.com",
    "mozilla.org",
    "w3c.org",
    "stackexchange.com",
    "stackoverflow.com",
    "serverfault.com",
    "superuser.com"
}

BLACK_LIST_PREFISES = {
    "cdn"
}
